/*

Licensed under the GNU AFFERO GENERAL PUBLIC LICENSE Version 3 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/agpl-3.0.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package application

import (
	"context"

	"github.com/presslabs/controller-util/syncer"

	settings "k8s.libre.sh/application/settings"
	"k8s.libre.sh/application/settings/parameters"
	"k8s.libre.sh/interfaces"
	"k8s.libre.sh/meta"
	"k8s.libre.sh/objects"
)

type Instance interface {
	meta.Instance
	GetComponents() map[int]Component
	GetOwner() interfaces.Object
	GetSettings() map[string]settings.Settings
	SetDefaults()
}

// TODO Transform to createandinit and return application object so we can store the values
func Init(i Instance, r interfaces.Reconcile) error {

	i.SetDefaults()

	setts, err := CreateAndInitSettings(i, r)
	if err != nil {
		return err
	}

	for _, c := range i.GetComponents() {

		c.Init()

		// TODO TOFIX
		component := c.GetComponent()

		meta.SetObjectMeta(i, c)

		if len(component) > 0 {
			c.SetComponent(component)
		}

		for _, o := range c.GetObjects() {
			meta.SetObjectMeta(c, o)
		}

		deps := c.HasSettings()

		for _, dep := range deps {
			foo := c.GetDependencies()
			foo = settings.MergeSettings(setts[dep].GetConfig(), foo)
		}

		/* 		if c.HasDependencies() {
			//	deps := c.GetDependencies()
			//	deps = settings.MergeSettings(sett.GetConfig(), deps)

		} */
		// we need setDefaults at last, objects should be init first and meta set
		c.SetDefaults()

	}

	return nil
}

func CreateAndInitSettings(i Instance, r interfaces.Reconcile) (setts map[string]settings.Settings, err error) {
	setts = i.GetSettings()

	values := make(map[string]interface{})
	val := make(map[string]interface{})

	for _, sett := range setts {
		err = sett.Init(r.GetClient())

		if err != nil {
			return setts, err
		}

		//	sett.SetDefaults()

		val[sett.GetMeta().GetComponent()] = parameters.KeyPairValues(sett.GetConfig().GetParameters())
		// TODO TOFIX should be run with all settings
		// 	sett.GetConfig().InitTemplateValues(values)
	}

	values["settings"] = val
	for _, sett := range setts {
		sett.GetConfig().InitTemplateValues(values)
	}

	return setts, err
}

func NewSyncers(app Instance, r interfaces.Reconcile, owner interfaces.Object) (syncers []syncer.Interface, err error) {

	////////////////////
	//// Settings //////
	///////////////////

	sett := app.GetSettings()
	/*
		err = sett.Init(r.GetClient())

		if err != nil {
			return syncers, err
		} */

	// Object Syncers

	/* 	for _, obj := range sett.GetObjects() {
	   		//	meta.SetObjectMeta(sett.CommonMeta, obj)
	   		s := objects.NewObjectSyncer(obj, owner, r)

	   		if err := syncer.Sync(context.TODO(), s, r.GetRecorder()); err != nil {
	   			return syncers, err
	   		}

	   	}
	*/
	for _, s := range sett {
		for _, obj := range s.GetObjects() {
			//	meta.SetObjectMeta(sett.CommonMeta, obj)
			objSyncer := objects.NewObjectSyncer(obj, owner, r)

			if err := syncer.Sync(context.TODO(), objSyncer, r.GetRecorder()); err != nil {
				return syncers, err
			}
		}

	}

	/////////////////////
	//// Components ////
	///////////////////
	components := app.GetComponents()

	for _, c := range components {
		/* 		if c.HasDependencies() {

			deps := c.GetDependencies()
			deps = settings.MergeSettings(sett, deps)

		} */
		// GET SYNCERS
		mutators := c.GetObjects()
		for _, m := range mutators {
			syncers = append(syncers, objects.NewObjectSyncer(m, owner, r))
		}
	}
	return syncers, nil

}
